### Persephone - a socio-economic-ecological model of European agricultural landscapes.
###
### This file includes functions for configuring the model and reading in map files.
###

## Note: much of this code was adapted from the GeMM model by Leidinger et al.
## (https://github.com/CCTB-Ecomods/gemm/blob/master/src/input.jl)

"""
    @param(domainparam)

Return a configuration parameter from the global settings.
The argument should be in the form `<domain>.<parameter>`,
for example `@param(core.outdir)`. Possible values for
<domain> are `core`, `nature`, `farm`, or `crop`. For a full
list of parameters, see `src/parameters.toml`.

Note that this macro only works in a context where the `model`
object is available.
"""
macro param(domainparam)
    domain = String(domainparam.args[1])
    paramname = String(domainparam.args[2].value)
    :($(esc(:model)).settings[$domain][$paramname])
end

"""
    getsettings(configfile, seed=nothing)

Combines all configuration options to produce a single settings dict.
Precedence: commandline parameters - user config file - default values
"""
function getsettings(configfile::String, seed::Union{Int64,Nothing}=nothing)
    # read in and merge configurations from the commandline, the default config file
    # and a user-supplied config file
    defaults = TOML.parsefile(configfile)
    commandline = parsecommandline()
    if haskey(commandline, "configfile") && isfile(commandline["configfile"])
        configs = TOML.parsefile(commandline["configfile"])
    else
        configs = nothing
    end
    settings = deepcopy(defaults)
    for domain in keys(defaults)
        for param in keys(defaults[domain])
            if param in keys(commandline)
                settings[domain][param] = commandline[param]
            elseif !isnothing(configs) && param in keys(configs[domain])
                settings[domain][param] = configs[domain][param]
            end
        end
    end
    # pre-process certain parameters
    if !isnothing(seed)
        settings["core"]["seed"] = seed
    elseif settings["core"]["seed"] == 0
        settings["core"]["seed"] = abs(rand(RandomDevice(), Int32))
    end
    defaultoutdir = defaults["core"]["outdir"]
    if settings["core"]["outdir"] == defaultoutdir
        outdir = defaultoutdir*"_"*string(Dates.today())*"_s"*string(settings["core"]["seed"])
        settings["core"]["outdir"] = outdir
    end
    if settings["core"]["startdate"] > settings["core"]["enddate"]
        Base.error("Enddate is earlier than startdate.")
    end
    settings
end

"""
    parsecommandline()

Certain software parameters can be set via the commandline.
"""
function parsecommandline()
    versionstring = """
            Persephone $(@project_version), commit $(read(`git rev-parse HEAD`, String)[1:8])
            © 2022-2023 Daniel Vedder, Lea Kolb (MIT license)
            https://git.idiv.de/xo30xoqa/persephone
            """
    s = ArgParseSettings(add_version=true, version=versionstring)
    @add_arg_table! s begin
        "--configfile", "-c"
            help = "name of the configuration file"
            arg_type = String
            required = false
        "--seed", "-s"
            help = "inital random seed"
            arg_type = Int
        "--outdir", "-o"
            help = "location of the output directory"
            arg_type = String
            required = false
        "--loglevel", "-l"
            help = "verbosity: \"debug\", \"info\", or \"quiet\""
            arg_type = String
            required = false
    end
    args = parse_args(s)
    for a in keys(args)
        (args[a] == nothing) && delete!(args, a)
    end
    args
end

